function getNativeConstructor(ext) {
  return ext ? document.createElement(ext).constructor : HTMLElement;
}

function extend(Base, init) {
  function Class(a, b, c) {
    // eslint-disable-next-line fp/no-this
    const el = this;

    if (!(el instanceof Class)) {
      return new Class(a, b, c);
    }

    const supr = () => {
      return typeof Reflect === 'undefined'
        ? Base.call(el)
        : Reflect.construct(Base, [], el.constructor);
    };

    return init.call(el, supr, a, b, c);
  }

  Class.prototype = Object.create(Base.prototype);
  Class.prototype.constructor = Class;

  return Class;
}

/**
 * Defines a custom element in the `CustomElementRegistry` which renders the component which is passed as an argument.
 *
 * @param {string} name - The tag name for the custom element.
 * @param {Function} [enhancer] - The element enhancer. You may optionally specify it to enhance the element with third-party capabilities such as middleware, custom renderer, public API, etc. The only element enhancers that ship with Swiss Element are `applyMiddleware` and `renderer`.
 * @param {Object|Array} [options] - Options object.
 * @param {Array} [options.observedAttributes] - Attributes to observe for adding, removing or changing which will trigger a component update if needed.
 * @param {string} [options.extends] - Specifies the built-in element your element inherits from if any (e.g. `extends: 'button'`).
 *
 * The options object is also passed to all the enhancers.
 *
 * @return {HTMLElement}
 */
function element(name, enhancer, options) {
  const Native = getNativeConstructor(options.extends);
  const SwissElement = extend(Native, function(supr, a, b, c) {
    return enhancer(supr)(options, a, b, c);
  });

  // Callbacks have to be on the prototype before construction.
  forwardCallbacks(SwissElement.prototype, [
    'connected',
    'disconnected',
    'attributeChanged',
    'adopted'
  ]);

  // `observedAttributes` have to be on the Class before construction.
  const oa = options.observedAttributes || [];
  options.observedAttributes = oa;
  SwissElement.observedAttributes = oa;
  SwissElement.toString = () => name;

  customElements.define(name, SwissElement, options);

  return SwissElement;
}

function forwardCallbacks(proto, callbacks) {
  callbacks.forEach(name => {
    const cb = name + 'Callback';
    let dispatching;
    proto[cb] = function(...args) {
      if (dispatching) return;

      // eslint-disable-next-line fp/no-this
      const el = this;
      dispatching = true;
      el[cb](...args);
      dispatching = false;
    };
  });
}

const isArray = Array.isArray;

function isString(value) {
  return typeof value === 'string';
}

function isFunction(value) {
  return typeof value === 'function';
}

function isUndefined(value) {
  return typeof value === 'undefined';
}

function compose(...fns) {
  return x => fns.filter(Boolean).reduceRight((y, f) => f(y), x);
}

function camelCase(name) {
  return name.replace(/-([a-z])/g, ($0, $1) => $1.toUpperCase());
}

function extend$1(target, ...sources) {
  sources.filter(Boolean).forEach((source) => {
    for (const key in source) { // eslint-disable-line fp/no-loops
      if (isFunction(target[key]) && isFunction(source[key])) {
        source[key].supr = target[key];
      }
      target[key] = source[key];
    }
  });
  return target;
}

function append(parent, nodes) {
  return []
    .concat(nodes)
    .map(node =>
      parent.appendChild(
        node instanceof Node ? node : document.createTextNode('' + node)
      )
    );
}

let idx = 0;
function uniqueId(prefix) {
  const id = `${prefix}${++idx}`;
  return customElements.get(id) ? uniqueId(prefix) : id;
}

function renderer(root, html, old) {
  if (html !== old) {
    if (isString(html)) {
      root.innerHTML = html;
    } else {
      root.innerHTML = '';
      append(root, html);
    }
  }
}

function componentEnhancer(component) {
  return createElement => (options, a, b, c) => {
    const comp = component || options.component;

    const el = createElement(options, a, b, c);
    let oldHtml;

    if (options.shadow) {
      el.attachShadow({ mode: options.shadow });
    }

    function connectedCallback() {
      el.update();
    }

    function attributeChangedCallback(name, oldValue, newValue) {
      if (el.shouldUpdate(name, oldValue, newValue)) {
        el.update();
      }
    }

    function shouldUpdate(name, oldValue, newValue) {
      return oldValue !== newValue;
    }

    function update(props) {
      if (props) extend$1(el, props);

      // If `renderer` is defined, the generated html is passed to `render`.
      // This is for elements that don't mutate the element in the component.
      el.render(el.renderer && comp.call(el, el));
    }

    function render(html) {
      if (isUndefined(html)) {
        // By default the html is generated in the `render` function.
        // This is for components that mutate the element in the component.
        html = comp.call(el, el);
      }

      const renderer$1 = el.renderer || renderer;
      // Leave render debouncing up to the render library.
      renderer$1(el.renderRoot(), html, oldHtml);
      oldHtml = html;
    }

    function renderRoot() {
      return el.shadowRoot || el._shadowRoot || el;
    }

    return extend$1(el, {
      connectedCallback,
      attributeChangedCallback,
      shouldUpdate,
      update,
      render,
      renderRoot
    });
  };
}

function propsToAttrs() {
  return createElement => (options, props, b, c) => {
    const el = createElement(options, props, b, c);
    const proto = Object.getPrototypeOf(el);
    const properties = {};

    options.observedAttributes.forEach(name => {
      const propName = camelCase(name);
      if (!(propName in proto)) {
        // eslint-disable-next-line fp/no-mutating-methods
        Object.defineProperty(proto, propName, {
          configurable: true,
          enumerable: true,
          get() {
            // eslint-disable-next-line fp/no-this
            return properties[propName] || this.getAttribute(name);
          },
          set(value) {
            properties[propName] = value;

            if (value == null) {
              // eslint-disable-next-line fp/no-this
              this.removeAttribute(name);
            } else {
              // Convert arrays and objects.
              if (typeof value === 'object') {
                try {
                  value = JSON.stringify(value);
                } catch (O_o) {
                  // At least we tried.
                }
              }
              // eslint-disable-next-line fp/no-this
              this.setAttribute(name, value);
            }
          }
        });
      }
    });

    return extend$1(el, props);
  };
}

/**
 * Quick and dirty way to add default enhancers and options.
 * @type {Object}
 * @ignore
 */
const options = {
  enhancers: [propsToAttrs(), componentEnhancer()]
};

/**
 * Defines a custom element in the `CustomElementRegistry` which renders the component which is passed as an argument.
 *
 * @param {string} name - The tag name for the custom element.
 * @param {Function} comp - The component that is rendered in the element.
 * @param {Function} [enhancer] - The element enhancer. You may optionally specify it to enhance the element with third-party capabilities such as middleware, custom renderer, public API, etc. The only element enhancers that ship with Swiss are `applyMiddleware` and `renderer`.
 * @param {Object|Array} [options] - Options object or `observedAttributes` only array for shorter syntax.
 * @param {Array} [options.observedAttributes] - Attributes to observe for adding, removing or changing which will trigger a component update if needed.
 * @param {string} [options.extends] - Specifies the built-in element your element inherits from if any (e.g. `extends: 'button'`).
 * @param {('open'|'closed')} [options.shadow] - Defines the shadow root mode, by default no shadow root is created and everything is rendered straight on the custom element.
 *
 * The options object is also passed to all the enhancers.
 *
 * @return {HTMLElement}
 */
function element$1(name, component, enhancer, opts) {
  if (isFunction(name)) {
    opts = enhancer;
    enhancer = component;
    component = name;
    name = uniqueId('s-');
  }

  if (!isFunction(enhancer) && isUndefined(opts)) {
    opts = enhancer;
    enhancer = undefined;
  }

  if (!isUndefined(enhancer) && !isFunction(enhancer)) {
    throw new Error('Enhancer should be a function.');
  }

  // To shorten syntax if opts is an array assume it's `observedAttributes`.
  if (isArray(opts)) {
    opts = { observedAttributes: opts };
  }

  opts = extend$1({ component }, options, opts);
  enhancer = compose(
    enhancer,
    ...options.enhancers
  );

  return element(name, enhancer, opts);
}

/**
 * Adds a simple way to define your own renderer.
 * Verified libraries working by passing just the `render` or `patch` function:
 *
 * - Lighterhtml
 * - Lit-html
 * - HTM-Preact
 * - Superfine
 *
 * @param  {Function} customRenderer A function that takes the custom element root and a function `html` which once executed renders the created dom nodes to the root node of the custom element.
 *
 * @return {Function}
 */
function rendererEnhancer(customRenderer = renderer) {
  return createElement => (...args) => {
    const element = createElement(...args);
    let render;
    let renderWays = [
      // default
      (root, html, old) => customRenderer(root, html, old),
      // lit-html, htm-preact
      (root, html, old) => customRenderer(html, root, old),
      // superfine
      (root, html, old) => customRenderer(old, html, root),
      // lighterhtml
      (root, html, old) => customRenderer(root, () => html, old)
    ];

    /**
     * Most library render functions look very similar, do a quick search on the
     * first render. Probably shouldn't do this but it's so damn convenient :P
     *
     * This function is only called on the first render pass, after it's cached.
     *
     * @param  {HTMLElement} root
     * @param  {Node|Function} html
     * @param  {Number} i
     * @return {*}
     */
    function renderer(root, html, old) {
      if (render) return render(root, html, old);

      render = renderWays[0];
      renderWays = renderWays.slice(1);

      let result;
      try {
        result = render(root, html, old);
      } catch (err) {
        if (renderWays.length > 0) {
          render = null;
          return renderer(root, html, old);
        }
      }

      return result || '';
    }

    return extend$1(element, {
      renderer
    });
  };
}

/**
 * Middleware lets you wrap the element's render method for fun and profit. The key feature of middleware is that it is composable. Multiple middleware can be combined together, where each middleware requires no knowledge of what comes before or after it in the chain.
 *
 * @param  {...Function} middleware Functions that conform to the Swiss _middleware_ API. Each middleware receives `SwissElement`'s `render` function as a named argument, and returns a function. That function will be given the `next` middleware's render method, and is expected to return a function of `fragment` calling `next(fragment)` with a potentially different argument, or at a different time, or maybe not calling it at all. The last middleware in the chain will receive the real element's `render` method as the `next` parameter, thus ending the chain. So, the middleware signature is `({ render }) => next => fragment`.
 *
 * @return {Function}
 */
function applyMiddleware(...middleware) {
  return createElement => (...args) => {
    const element = createElement(...args);
    let render = () => {
      throw new Error('Middleware should not render in setup.');
    };

    const middlewareAPI = {
      nodeName: element.nodeName,
      render: html => render(html),
      getTree: () => element.outerHTML
    };

    const chain = middleware.map(mw => mw(middlewareAPI));
    render = compose(...chain)(element.render);

    return extend$1(element, {
      render
    });
  };
}

export { applyMiddleware, componentEnhancer as component, compose, element$1 as element, extend$1 as extend, options, propsToAttrs, rendererEnhancer as renderer };
